/*
 *  * Copyright © 2019 Pendo.io. All rights reserved.
 */

import 'dart:async';
import 'dart:io' show Platform;
import 'package:flutter/widgets.dart';
// ignore: prefer_relative_imports
import 'package:pendo_sdk/src/channel.dart';
// ignore: prefer_relative_imports
import 'package:pendo_sdk/src/common/consts.dart';
// ignore: prefer_relative_imports
import 'package:pendo_sdk/src/debouncer.dart';
// ignore: prefer_relative_imports
import 'package:pendo_sdk/src/logger.dart';
// ignore: prefer_relative_imports
import 'package:pendo_sdk/src/pendo_scan.dart';
import 'src/exstension/map_extension.dart';

export 'package:pendo_sdk/src/pendo_navigation_observer.dart' show PendoNavigationObserver;

class _PendoSDKConsts {
  static const _kTrack = 'Track';
  static const _kObservable = 'Observable';
  static const _kIntegrationType = 'IntegrationType';
  static const _debouncerDuration = Duration(milliseconds: 110);
}

typedef PendoFlutterPlugin = PendoSDK;

class PendoSDK {
  static const String FILE_TAG = 'PendoSDK:';
  static VoidCallback? _onFrameDataChangeOriginal;
  static bool _isScanningCallbackInitialized = false;
  static final PendoDebouncer _debouncer = PendoDebouncer(duration: _PendoSDKConsts._debouncerDuration);

  /// Setups the Pendo SDK.
  ///
  /// Use the [appKey] to setup the SDK.
  /// Setting up the SDK <b>does not</b> start a session.
  ///
  /// [pendoOptions] is for internal use and is optional.
  ///
  /// Example usage:
  /// `PendoSDK.setup('YOUR_PENDOConsts.kEY');`
  ///
  /// Call the `startSession` API when you would like to start tracking analytics and/or presenting guides (for anonymous or identified visitors).
  static Future<void> setup(String appKey, {Map<String, dynamic>? pendoOptions}) async {
    WidgetsFlutterBinding.ensureInitialized();

    if (!_isScanningCallbackInitialized && pendoOptions != null) {
      var integrationType = pendoOptions[_PendoSDKConsts._kIntegrationType];

      if (integrationType == _PendoSDKConsts._kObservable) {
        Logger.i('$FILE_TAG Pendo Flutter Observable type detected');
        _isScanningCallbackInitialized = true;
        _subscribeToScanningCallback();
      } else {
        pendoOptions[_PendoSDKConsts._kIntegrationType] = _PendoSDKConsts._kTrack;
      }
    }

    try {
      var arguments = {
        Consts.kApplicationKey: appKey,
        Consts.kPendoOptions: pendoOptions,
        Consts.kPlatformVersion: _flutterVersion(),
      };

      return await PendoChannel.invokeMethod(Consts.kSetup, arguments);
    } catch (e) {
      throw 'Failed calling $Consts.kSetup() with $Consts.kApplicationKey: $appKey, $Consts.kPendoOptions: $pendoOptions, error is: $e';
    }
  }

  /// Must be called <b>after</b> the SDK was setup.
  ///
  /// Starts a session with a new visitor.
  /// In case a session was already started before, end the previous one and start a new one.
  static Future<void> startSession(String? visitorId, String? accountId, Map<String, dynamic>? visitorData, Map<String, dynamic>? accountData) async {
    WidgetsFlutterBinding.ensureInitialized();

    try {
      var arguments = {
        Consts.kVisitorID: visitorId,
        Consts.kAccountID: accountId,
        Consts.kVisitorData: visitorData?.validateData(),
        Consts.kAccountData: accountData?.validateData()
      };

      return await PendoChannel.invokeMethod(Consts.kStartSession, arguments);
    } catch (e) {
      throw 'Failed calling $Consts.kStartSession() with $Consts.kVisitorID: $visitorId, $Consts.kAccountID: $accountId, $Consts.kVisitorData: $visitorData, $Consts.kAccountData: $accountData, error is: $e';
    }
  }

  /// Sets a visitor data value.
  ///
  /// This data is used by Pendo Mobile for creating audiences or reporting analytics.
  ///
  /// For instance you might want to provide data on the visitor's age or if the visitor is logged into a service.
  static Future<void> setVisitorData(Map<String, dynamic>? visitorData) async {
    WidgetsFlutterBinding.ensureInitialized();

    try {
      var arguments = {Consts.kVisitorData: visitorData?.validateData()};
      return await PendoChannel.invokeMethod(Consts.kSetVisitorData, arguments);
    } catch (e) {
      throw 'Failed calling $Consts.kSetVisitorData() with $Consts.kVisitorData: $visitorData, error is: $e';
    }
  }

  /// Sets account data value for a given data name.
  ///
  /// This data is used by Pendo Mobile for creating audiences or reporting analytics.
  ///
  /// For instance you might want to provide data on the account's subscription or if the account is active or not.
  static Future<void> setAccountData(Map<String, dynamic>? accountData) async {
    WidgetsFlutterBinding.ensureInitialized();

    try {
      var arguments = {Consts.kAccountData: accountData?.validateData()};
      return await PendoChannel.invokeMethod(Consts.kSetAccountData, arguments);
    } catch (e) {
      throw 'Failed calling $Consts.kSetAccountData() with $Consts.kAccountData: $accountData, error is: $e';
    }
  }

  /// Tracks an event that has happened in your application.
  static Future<void> track(String event, Map<String, dynamic>? properties) async {
    WidgetsFlutterBinding.ensureInitialized();

    try {
      var arguments = {
        Consts.kTrackEventName: event,
        Consts.kTrackEventProperties: properties?.validateData()
      };
      return await PendoChannel.invokeMethod(Consts.kTrack, arguments);
    } catch (e) {
      throw 'Failed calling $Consts.kTrack() with $Consts.kTrackEventName: $event, $Consts.kTrackEventProperties: $properties, error is: $e';
    }
  }

  /// Must be called <b>after</b> the SDK was setup.
  ///
  /// Ends the current session. As a result analytics gathering and presentation of guides will stop.
  static Future<void> endSession() async {
    WidgetsFlutterBinding.ensureInitialized();

    try {
      return await PendoChannel.invokeMethod(Consts.kEndSession);
    } catch (e) {
      throw 'Failed calling $Consts.kEndSession(), error is: $e';
    }
  }

  /// Stops showing guides.
  ///
  /// Setting [dismissGuides] to true will close any guide currently presented on the screen.
  static Future<void> pauseGuides({bool dismissGuides = true}) async {
    WidgetsFlutterBinding.ensureInitialized();

    try {
      Map<String, bool> arguments = {Consts.kDismissGuides: dismissGuides};
      return await PendoChannel.invokeMethod(Consts.kPauseGuides, arguments);
    } catch (e) {
      throw 'Failed calling $Consts.kPauseGuides() with $Consts.kDismissGuides: $dismissGuides, error is: $e';
    }
  }

  /// Dismisses any visible guides.
  static Future<void> dismissVisibleGuides() async {
    WidgetsFlutterBinding.ensureInitialized();

    try {
      return await PendoChannel.invokeMethod(Consts.kDismissVisibleGuides);
    } catch (e) {
      throw 'Failed calling $Consts.kDismissVisibleGuides(), error is: $e';
    }
  }

  /// Resumes presenting guides.
  static Future<void> resumeGuides() async {
    WidgetsFlutterBinding.ensureInitialized();

    try {
      return await PendoChannel.invokeMethod(Consts.kResumeGuides);
    } catch (e) {
      throw 'Failed calling $Consts.kResumeGuides(), error is: $e';
    }
  }

  /// Returns the current visitor id if available.
  static Future<String?> getVisitorId() async {
    WidgetsFlutterBinding.ensureInitialized();

    try {
      return await PendoChannel.invokeMethod(Consts.kGetVisitorId);
    } catch (e) {
      throw 'Failed calling $Consts.kGetVisitorId(), error is: $e';
    }
  }

  /// Returns the current account id if available.
  static Future<String?> getAccountId() async {
    WidgetsFlutterBinding.ensureInitialized();

    try {
      return await PendoChannel.invokeMethod(Consts.kGetAccountId);
    } catch (e) {
      throw 'Failed calling $Consts.kGetAccountId(), error is: $e';
    }
  }

  /// Returns the Pendo SDK unique device id.
  ///
  /// Used for anonymous visitor id.
  /// This id is unique per application.
  static Future<String?> getDeviceId() async {
    WidgetsFlutterBinding.ensureInitialized();
    
    try {
      return await PendoChannel.invokeMethod(Consts.kGetDeviceId);
    } catch (e) {
      throw 'Failed calling $Consts.kGetDeviceId(), error is: $e';
    }
  }

  /// Presents debug logs when [isDebugEnabled] is set to true.
  static Future<void> setDebugMode({bool isDebugEnabled = true}) async {
    WidgetsFlutterBinding.ensureInitialized();

    try {
      if (isDebugEnabled) {
        Logger.debugEnabled();
      }
      Map<String, bool> arguments = {Consts.kIsDebugEnabled: isDebugEnabled};
      return await PendoChannel.invokeMethod(Consts.kSetDebugMode, arguments);
    } catch (e) {
      throw 'Failed calling $Consts.kSetDebugMode() with $Consts.kIsDebugEnabled: $isDebugEnabled, error is: $e';
    }
  }

  // Private
  /// Initiates a callback to scan for internal navigation types according to [pendoOptions] passed in setup call.
  static void _subscribeToScanningCallback() {
    Logger.d('$FILE_TAG Subscribing to scanning callback...');

    try {
      _onFrameDataChangeOriginal = WidgetsBinding.instance.platformDispatcher.onFrameDataChanged;

      WidgetsBinding.instance.platformDispatcher.onFrameDataChanged = () {
        _debouncer.execute(() {
          _onFrameDataChangeOriginal?.call();
          Logger.i('$FILE_TAG onFrameDataChanged invoked');
          PendoScan().scanFrom(WidgetsBinding.instance.rootElement);
        });
      };
    } catch (e) {
      Logger.f('$FILE_TAG couldn\'t subscribe to onFrameDataChanged. exception: $e');
    }
  }

  static String _flutterVersion() => Platform.version.split(' ')[0];
}
