/*
 *  * Copyright © 2019 Pendo.io. All rights reserved.
 */

import 'dart:async';
import 'package:collection/collection.dart';
import 'package:flutter/material.dart';
// ignore: prefer_relative_imports
import 'package:pendo_sdk/pendo_sdk.dart';
// ignore: prefer_relative_imports
import 'package:pendo_sdk/src/channel.dart';
// ignore: prefer_relative_imports
import 'package:pendo_sdk/src/common/consts.dart';
// ignore: prefer_relative_imports
import 'package:pendo_sdk/src/logger.dart';

class _PendoScanConsts {
  // element types
  static const _kBottomNavigationBar = 'BottomNavigationBar';
  static const _kTabBar = 'TabBar';
  static const _kDrawer = 'Drawer';
  static const _kEndDrawer = 'EndDrawer';
  // keys
  static const _kRetroElementInfo = 'retroElementInfo';
  static const _kElementType = 'elementType';
  static const _kElementName = 'elementName';
  static const _kDrawerState = 'drawerState';
  static const _kDrawerClose = 'close';
  static const _kDrawerOpen = 'open';
  static const _kIndex = 'index';
  static const _kTotal = 'total';
  static const _kDepth = 'depth';
}

class PendoScan {
  static const FILE_TAG = 'PendoScan';

  /// last data
  Map lastInvokedArguments = {};

  /// Properties
  int _depth = 1;
  List _info = [];

  // for tests usage
  List get info => _info;
  void clearInfo() => _info.clear();
  void resetDepth() => _depth = 1;

  // singleton, mainly for tests usage
  PendoScan._internal();
  factory PendoScan() => _singleton;
  static final PendoScan _singleton = PendoScan._internal();

  /// API
  void scanFrom(Element? element) {
    if (element != null && element.mounted) {
      Logger.i('$FILE_TAG scan started....');
      final timeNow = DateTime.now();
      visitChild(element);
      final difference = DateTime.now().difference(timeNow);
      Logger.i('$FILE_TAG scan finished in $difference');
    } else {
      Logger.c('$FILE_TAG element not mounted. dropping internal scan');
    }

    notifySDK();
  }

  /// Private use
  void visitChild(Element element) {
    try {
      if (element.widget is BottomNavigationBar) {
        extractBottomNavigationBar(element.widget as BottomNavigationBar);
      } else if (element.widget is TabBar) {
        extractTabBar(element.widget as TabBar);
      } else if (element.widget is Scaffold && element is StatefulElement && element.state is ScaffoldState) {
        ScaffoldState scaffoldState = element.state as ScaffoldState;

        if (scaffoldState.isDrawerOpen) {
          extractDrawer(scaffoldState);
        }

        if (scaffoldState.isEndDrawerOpen) {
          extractEndDrawer(scaffoldState);
        }
      }

      element.visitChildren(visitChild);
    } catch (e) {
      Logger.f('$FILE_TAG couldn\'t visit child $element, exception caught: $e');
    }
  }

  void extractBottomNavigationBar(BottomNavigationBar bottomNavigationBar) {
    if (bottomNavigationBar.items.length > 0) {
      int index = bottomNavigationBar.currentIndex;
      int length = bottomNavigationBar.items.length;
      String? barName = bottomNavigationBar.items.elementAt(index).label;

      final dynamic infoElement = {
        _PendoScanConsts._kRetroElementInfo: {
          _PendoScanConsts._kElementType:
          _PendoScanConsts._kBottomNavigationBar,
          _PendoScanConsts._kElementName: (barName != null ? barName : ''),
          _PendoScanConsts._kIndex: index + 1,
          _PendoScanConsts._kTotal: length,
          _PendoScanConsts._kDepth: _depth++
        }
      };

      _info.add(infoElement);
    }
  }

  void extractTabBar(TabBar tabBarWidget) {
    var tabBarController = tabBarWidget.controller;

    if (tabBarController != null && tabBarController.length > 0) {
      int index = tabBarController.index;
      int length = tabBarController.length;
      String? tabName = (tabBarWidget.tabs.elementAt(index) as Tab).text;

      final dynamic infoElement = {
        _PendoScanConsts._kRetroElementInfo: {
        _PendoScanConsts._kElementType: _PendoScanConsts._kTabBar,
        _PendoScanConsts._kElementName: (tabName != null ? tabName : ''),
        _PendoScanConsts._kIndex: index + 1,
        _PendoScanConsts._kTotal: length,
        _PendoScanConsts._kDepth: _depth++
        }
      };

      _info.add(infoElement);
    }
  }

  void extractDrawer(ScaffoldState scaffoldState) {
    final dynamic infoElement = {
      _PendoScanConsts._kRetroElementInfo: {
        _PendoScanConsts._kElementType: _PendoScanConsts._kDrawer,
        _PendoScanConsts._kDrawerState: scaffoldState.isDrawerOpen ? _PendoScanConsts._kDrawerOpen : _PendoScanConsts._kDrawerClose,
        _PendoScanConsts._kDepth: _depth++
      }
    };

    _info.add(infoElement);
  }

  void extractEndDrawer(ScaffoldState scaffoldState) {
    final dynamic infoElement = {
      _PendoScanConsts._kRetroElementInfo: {
        _PendoScanConsts._kElementType: _PendoScanConsts._kEndDrawer,
        _PendoScanConsts._kDrawerState: scaffoldState.isEndDrawerOpen ? _PendoScanConsts._kDrawerOpen : _PendoScanConsts._kDrawerClose,
        _PendoScanConsts._kDepth: _depth++
      }
    };

    _info.add(infoElement);
  }

  void notifySDK() {
    String? screenName = (PendoNavigationObserver.routeName != null ? PendoNavigationObserver.routeName : '');
    final info = _info.length > 0 ? _info[0] : {};
    final arguments = {
      Consts.kRetroactiveScreenId: screenName,
      Consts.kInfo: info
    };

    _info.clear();
    _depth = 1;

    if (DeepCollectionEquality().equals(arguments, lastInvokedArguments)) {
      Logger.c('$FILE_TAG dropping scan due to equal arguments for $screenName');
      return;
    }

    lastInvokedArguments = arguments;
    Logger.d('$FILE_TAG page changed with: $arguments');
    unawaited(PendoChannel.invokeMethod(Consts.kScreenChanged, arguments));
  }
}
